use std::{error::Error, sync::Arc};

use vulkano::{
    command_buffer::{
        AutoCommandBufferBuilder, PrimaryAutoCommandBuffer, RenderPassBeginInfo, SubpassBeginInfo,
        SubpassEndInfo,
    },
    descriptor_set::DescriptorSet,
    pipeline::{GraphicsPipeline, Pipeline},
    render_pass::{Framebuffer, RenderPass},
};

pub trait AsRecordable<T> {
    fn record_render_commands(
        builder: &mut AutoCommandBufferBuilder<PrimaryAutoCommandBuffer>,
        data: &T,
        pipeline: &Arc<GraphicsPipeline>,
        descriptor_sets: &[Arc<DescriptorSet>],
    ) -> Result<(), Box<dyn Error>>;
}

pub trait AsDrawable<T> {
    type VertexBuffer;
    type IndexBuffer;

    fn vertex_buffer(data: &T) -> &Self::VertexBuffer;

    fn index_buffer(_data: &T) -> Option<&Self::IndexBuffer> {
        None
    }

    fn vertex_count(data: &T) -> u32;

    fn index_count(_data: &T) -> u32 {
        0
    }

    fn instance_count(_data: &T) -> u32 {
        1
    }

    fn first_vertex(_data: &T) -> u32 {
        0
    }

    fn first_index(_data: &T) -> u32 {
        0
    }

    fn first_instance(_data: &T) -> u32 {
        0
    }
}

pub trait AsRenderPassRecordable<T> {
    fn begin_render_pass(
        builder: &mut AutoCommandBufferBuilder<PrimaryAutoCommandBuffer>,
        _render_pass: &Arc<RenderPass>,
        framebuffer: &Arc<Framebuffer>,
        clear_values: Vec<Option<vulkano::format::ClearValue>>,
    ) -> Result<(), Box<dyn Error>> {
        builder.begin_render_pass(
            RenderPassBeginInfo {
                clear_values,
                ..RenderPassBeginInfo::framebuffer(framebuffer.clone())
            },
            SubpassBeginInfo {
                contents: vulkano::command_buffer::SubpassContents::Inline,
                ..Default::default()
            },
        )?;
        Ok(())
    }

    fn end_render_pass(
        builder: &mut AutoCommandBufferBuilder<PrimaryAutoCommandBuffer>,
    ) -> Result<(), Box<dyn Error>> {
        builder.end_render_pass(SubpassEndInfo::default())?;
        Ok(())
    }

    fn record_render_pass_commands(
        builder: &mut AutoCommandBufferBuilder<PrimaryAutoCommandBuffer>,
        data: &T,
    ) -> Result<(), Box<dyn Error>>;
}

pub trait AsRenderableObject<T>: AsRecordable<T> + AsDrawable<T> {
    fn record_complete_render(
        builder: &mut AutoCommandBufferBuilder<PrimaryAutoCommandBuffer>,
        data: &T,
        pipeline: &Arc<GraphicsPipeline>,
        descriptor_sets: &[Arc<DescriptorSet>],
    ) -> Result<(), Box<dyn Error>> {
        builder.bind_pipeline_graphics(pipeline.clone())?;

        if !descriptor_sets.is_empty() {
            builder.bind_descriptor_sets(
                vulkano::pipeline::PipelineBindPoint::Graphics,
                pipeline.layout().clone(),
                0,
                descriptor_sets.iter().cloned().collect::<Vec<_>>(),
            )?;
        }

        Self::record_render_commands(builder, data, pipeline, descriptor_sets)?;

        Ok(())
    }
}
