use std::sync::Arc;

use glam::{Mat4, Quat, Vec3};
use vulkano::{
    Validated,
    buffer::{AllocateBufferError, BufferContents, BufferCreateInfo, BufferUsage, Subbuffer},
    memory::allocator::{AllocationCreateInfo, MemoryTypeFilter, StandardMemoryAllocator},
    pipeline::graphics::vertex_input::Vertex,
};

use crate::core::render::primitives::{AsBindableBuffer, AsVertexBuffer};

pub struct Transform {
    pub position: Vec3,
    pub rotation: Quat,
    pub scale: Vec3,
}

#[derive(BufferContents, Vertex, Clone, Copy)]
#[repr(C)]
pub struct TransformRaw {
    #[format(R32G32B32A32_SFLOAT)]
    pub model: [[f32; 4]; 4],
}

impl Default for Transform {
    fn default() -> Self {
        Self {
            position: Vec3::default(),
            rotation: Quat::default(),
            scale: Vec3::ONE,
        }
    }
}

impl Transform {
    pub fn rotate(&mut self, rotation: Quat) {
        self.rotation *= rotation;
    }

    pub fn translate(&mut self, translation: Vec3) {
        self.position += translation;
    }

    pub fn scale(&mut self, scale: Vec3) {
        self.scale *= scale;
    }

    pub fn to_raw_tranform(&self) -> TransformRaw {
        TransformRaw {
            model: (Mat4::from_translation(self.position)
                * Mat4::from_quat(self.rotation)
                * Mat4::from_scale(self.scale))
            .to_cols_array_2d(),
        }
    }

    pub fn create_buffer(
        memory_allocator: &Arc<StandardMemoryAllocator>,
        transforms: &[Transform],
    ) -> Result<Subbuffer<[TransformRaw]>, Validated<AllocateBufferError>> {
        TransformRaw::create_vertex_buffer(
            memory_allocator,
            &transforms
                .iter()
                .map(|t| t.to_raw_tranform())
                .collect::<Vec<_>>(),
        )
    }
}

impl AsBindableBuffer<TransformRaw> for TransformRaw {
    type BufferData = TransformRaw;

    fn buffer_create_info() -> BufferCreateInfo {
        BufferCreateInfo {
            usage: BufferUsage::VERTEX_BUFFER,
            ..Default::default()
        }
    }

    fn allocation_create_info() -> AllocationCreateInfo {
        AllocationCreateInfo {
            memory_type_filter: MemoryTypeFilter::PREFER_DEVICE
                | MemoryTypeFilter::HOST_SEQUENTIAL_WRITE,
            ..Default::default()
        }
    }

    fn to_buffer_data(data: &TransformRaw) -> Self::BufferData {
        *data
    }
}

impl AsVertexBuffer<TransformRaw> for TransformRaw {}
